/*
 * term.c
 * terminal handling routines for cancan
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#ifdef USETERMIO
#ifdef APOLLO
#include "/sys5.3/usr/include/sys/termio.h"
#else
/*
 * including both termio.h and termios.h might be an overkill, and gives
 * many warnings, but seems to be necessary at times. works anyway.
 */
#include <termio.h>
#include <termios.h>
#endif
#else /* not USETERMIO */
#include <sys/ioctl.h>
#endif

#ifdef BSD_386
#include <sys/ioctl_compat.h>
#define O_ECHO ECHO
#define O_CBREAK CBREAK
#endif

#include "cancan.h"
#include "edit.h"


/* int ioctl(); */

#ifdef __hpux  /* cc for HP-UX  SHOULD define this... */
#define TCGETS TCGETATTR
#define TCSETS TCSETATTR
#endif



#ifndef VT100
/* termcap function declarations */
extern int tgetent();
extern int tgetnum();
extern int tgetflag();
extern char *tgetstr();
extern char *tgoto();
#endif

/* terminal escape sequenses */
char kpadstart[CAPLEN], kpadend[CAPLEN], begoln[CAPLEN], clreoln[CAPLEN],
     leftcur[CAPLEN], rightcur[CAPLEN], curgoto[CAPLEN], inschar[CAPLEN],
     delchar[CAPLEN], clreoscr[CAPLEN], upcur[CAPLEN];
/* attribute changers: */
char modenorm[CAPLEN], modebold[CAPLEN], modeinv[CAPLEN], modeuline[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN];

int wrapglitch;			/* strange wrap behaviour (a la vt100) */
static int gotocost;		/* length of a typical tgoto-string */
int metakey;			/* terminal has meta key setting high bit */

#ifdef USETERMIO
#ifdef TCSETS
struct termios ttybsave;
#else
struct termio ttybsave;
#endif
#else
struct sgttyb ttybsave;
struct tchars tcsave;
struct ltchars ltcsave;
#endif

static int extract();

/*
 * Terminal handling routines:
 * These are one big mess of left-justified chicken scratches.
 * It should be handled more cleanly...but unix portability is what it is.
 */

/*
 * Set the terminal to character-at-a-time-without-echo mode, and save the
 * original state in ttybsave
 */
void set_terminal()
{
#ifdef USETERMIO
#ifdef TCSETS
  struct termios ttyb;
  ioctl(0, TCGETS, &ttyb);
#else
  struct termio ttyb;
  ioctl(0, TCGETA, &ttyb);
#endif
  ttybsave = ttyb;
  ttyb.c_lflag &= ~(ECHO|ICANON);
  ttyb.c_cc[VTIME] = 0;
  ttyb.c_cc[VMIN] = 1;
  /* disable the special handling of the suspend key (handle it ourselves) */
#ifdef VSUSP
  ttyb.c_cc[VSUSP] = 0;
#else
#ifdef SWTCH
  ttyb.c_cc[SWTCH] = 0;
#else
  ttyb.c_cc[SUSP] = 0;
#endif
#endif
#ifdef TCSETS
  ioctl(0, TCSETS, &ttyb);
#else
  ioctl(0, TCSETA, &ttyb);
#endif
#else /* not USETERMIO */
  struct sgttyb ttyb;
  struct ltchars ltc;
  ioctl(0, TIOCGETP, &ttybsave);
  ioctl(0, TIOCGETC, &tcsave);
  ioctl(0, TIOCGLTC, &ltcsave);
  ttyb = ttybsave;
  ttyb.sg_flags = (ttyb.sg_flags|O_CBREAK) & ~O_ECHO;
  ioctl(0, TIOCSETP, &ttyb);
  ltc = ltcsave;
  ltc.t_suspc = -1;
  ioctl(0, TIOCSLTC, &ltc);
#endif
  printf("%s", kpadstart); fflush(stdout);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
#ifdef USETERMIO
#ifdef TCSETS
  ioctl(0, TCSETS, &ttybsave);
#else
  ioctl(0, TCSETA, &ttybsave);
#endif
#else
  ioctl(0, TIOCSETP, &ttybsave);
  ioctl(0, TIOCSLTC, &ltcsave);
#endif
  printf("%s", kpadend); fflush(stdout);
}

/*
 * enable/disable special keys depending on the current linemode
 */
void set_spec_keys()
{
#ifdef USETERMIO
  int i;
#ifdef TCSETS
  struct termios ttyb;
  ioctl(0, TCGETS, &ttyb);
#else
  struct termio ttyb;
  ioctl(0, TCGETA, &ttyb);
#endif
  if(linemode & LM_CHAR) 
   {
     /* char-by-char mode: disable all special keys and set raw mode */
     for(i = 0; i < NCCS; i++)
     	ttyb.c_cc[i] = 0;
     ttyb.c_oflag &= ~OPOST;
   }
  else 
   {
     /* line at a time mode: enable them, except suspend */
     for(i = 0; i < NCCS; i++)
     	ttyb.c_cc[i] = ttybsave.c_cc[i];
     /* disable the suspend key (handle it ourselves) */
#ifdef VSUSP
     ttyb.c_cc[VSUSP] = 0;
#else
#ifdef SWTCH
     ttyb.c_cc[SWTCH] = 0;
#else
     ttyb.c_cc[SUSP] = 0;
#endif
#endif
     /* set cooked mode */
     ttyb.c_oflag |= OPOST;
   }
#ifdef TCSETS
  ioctl(0, TCSETS, &ttyb);
#else
  ioctl(0, TCSETA, &ttyb);
#endif
#else /* not USETERMIO */
  struct tchars tc = {-1, -1, -1, -1, -1, -1};
  struct ltchars ltc = {-1, -1, -1, -1, -1, -1};
  struct sgttyb ttyb;
  ioctl(0, TIOCGETP, &ttyb);
  if(linemode & LM_CHAR) 
   {
     /* char-by-char mode: set RAW mode*/
     ttyb.sg_flags |= RAW;
   }
  else 
   {
     /* line-at-a-time mode: enable spec keys, disable RAW */
     tc = tcsave;
     ltc = ltcsave;
     ltc.t_suspc = -1;	/* suspend key remains disabled */
     ttyb.sg_flags &= ~RAW;
   }
  ioctl(0, TIOCSETP, &ttyb);
  ioctl(0, TIOCSETC, &tc);
  ioctl(0, TIOCSLTC, &ltc);
#endif
}

/*
 * read termcap definitions
 */
void settermcap()
{
#ifndef VT100
    char tcbuf[2048];		/* by convention, this is enough */
    char buf[CAPLEN];
    int v;
#endif    
#if !defined(VT100) || defined(TELNETBUG)
    char *term = getenv("TERM");
    if(!term) {
	printf("$TERM not set\n"); exitcancan();
    }
#endif
#ifdef VT100
    /* hard-coded vt100 features if no termcap: */
    strcpy(begoln, "\r");
    strcpy(clreoln, "\033[K");
    strcpy(clreoscr, "\033[J");
    strcpy(leftcur, "\033[D");
    strcpy(rightcur, "\033[C");
    strcpy(upcur, "\033[A");
    strcpy(modenorm, "\033[m");
    strcpy(modebold, "\033[1m");
    strcpy(modeinv, "\033[7m");
    strcpy(modeuline, "\033[4m");
    /* not all of the above attributes exist on all vt100s */
    cols = 80;
#ifdef LINES
    lines = LINES;
#else
    lines = 24;
#endif
    *kpadstart = *kpadend = 0;
    add_keynode("left", "\033[D", prev_char, NULL);
    add_keynode("right", "\033[C", next_char, NULL);
    add_keynode("up", "\033[A", prev_line, NULL);
    add_keynode("down", "\033[B", next_line, NULL);
    wrapglitch = 1;
    metakey = 0;		/* vt100 has no meta key */
    gotocost = 7;		/* average for vt100 cursor movement? */
#else /* not VT100 */
    v = tgetent(tcbuf, term);
    if(v < 0)
	syserr("tgetent");
    if(v == 0) {
	printf("There is no entry for '%s' in the terminal data base.\n",
	       term);
	printf("Please set your $TERM environment variable correctly.\n");
	exitcancan();
    }
    if(!extract("cr", begoln, 0)) strcpy(begoln, "\r");
    extract("ce", clreoln, 1);
    extract("cd", clreoscr, 1);
    if(!extract("le", leftcur, 0)) strcpy(leftcur, "\b");
    extract("nd", rightcur, 1);
    extract("up", upcur, 0);
    extract("cm", curgoto, 1);
    cols = tgetnum("co");
    lines = tgetnum("li");
    /* set inschar to 'im' + 'ic' + space + 'ei' + 'le' */
    /* this is inefficient, but better than nothing. */
    extract("im", inschar, 0);
    extract("ic", inschar + strlen(inschar), 0);
    if(strlen(inschar) > 0) {
	/* is there a way to insert character? */
	strcat(inschar, " ");
	extract("ei", inschar + strlen(inschar), 0);
	strcat(inschar, leftcur);
    }
    extract("dm", delchar, 0);
    extract("dc", delchar + strlen(delchar), 0);
    extract("ed", delchar + strlen(delchar), 0);
    extract("ks", kpadstart, 0);
    extract("ke", kpadend, 0);
    if(extract("kl", buf, 0)) add_keynode("left", buf, prev_char, NULL);
    if(extract("kr", buf, 0)) add_keynode("right", buf, next_char, NULL);
    if(extract("ku", buf, 0)) add_keynode("up", buf, prev_line, NULL);
    if(extract("kd", buf, 0)) add_keynode("down", buf, next_line, NULL);
    extract("me", modenorm, 0);
    extract("md", modebold, 0);
    extract("mr", modeinv, 0);
    extract("us", modeuline, 0);
    extract("so", modestandon, 0);
    extract("se", modestandoff, 0);
    gotocost = strlen(tgoto(curgoto, 0, 10));
    wrapglitch = tgetflag("xn");
    metakey = tgetflag("km");
#endif
#ifdef TELNETBUG
    if(strncmp(term, "vt10", 4) == 0) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;3%d;4%dm", DEFAULTFG, DEFAULTBG);
    }
#endif
}

/*
 * add the default keypad bindings to the list
 */
void default_keybindings()
{
    /*
     * Note: termcap doesn't have sequences for the numeric keypad, so we just
     * assume they are the same as for a vt100.
     */
    add_keynode("kp2", "\033Or", key_run_command, "s");
    add_keynode("kp3", "\033Os", key_run_command, "d");
    add_keynode("kp4", "\033Ot", key_run_command, "w");
    add_keynode("kp5", "\033Ou", key_run_command, "exits");
    add_keynode("kp6", "\033Ov", key_run_command, "e");
    add_keynode("kp8", "\033Ox", key_run_command, "n");
    add_keynode("kp9", "\033Oy", key_run_command, "u");
}
    
/*
 * initialize the key binding list
 */
void init_keybindings()
{
    static char buf[2] = "x";
    /* standard emacs-like key bindings */
    /* (should perhaps be sorted in decreasing frequency of use) */
    add_keynode("return", "\r", enter_line, NULL);
    add_keynode("lf", "\n", enter_line, NULL);
    add_keynode("tab", "\t", complete_word, NULL);
    add_keynode("bs", "\b", del_char_left, NULL);
    add_keynode("del", "\177", del_char_left, NULL);
    add_keynode("C-b", (*buf = CONTROL('b'), buf) , prev_char, NULL);
    add_keynode("C-f", (*buf = CONTROL('f'), buf) , next_char, NULL);
    add_keynode("C-p", (*buf = CONTROL('p'), buf) , prev_line, NULL);
    add_keynode("C-n", (*buf = CONTROL('n'), buf) , next_line, NULL);
    add_keynode("C-y", (*buf = CONTROL('y'), buf) , enter_into_history,	NULL);
    add_keynode("C-d", (*buf = CONTROL('d'), buf) , del_char_right, NULL);
    add_keynode("C-a", (*buf = CONTROL('a'), buf) , beginning_of_line, NULL);
    add_keynode("C-e", (*buf = CONTROL('e'), buf) , end_of_line, NULL);
    add_keynode("C-k", (*buf = CONTROL('k'), buf) , kill_to_eol, NULL);
    add_keynode("C-t", (*buf = CONTROL('t'), buf) , transpose_chars, NULL);
    add_keynode("C-l", (*buf = CONTROL('l'), buf) , redraw_input_line, NULL);
    add_keynode("C-z", (*buf = CONTROL('z'), buf) , suspend_cancan, NULL);
    add_keynode("M-d", "\033d", del_word_right, NULL);
    add_keynode("M-f", "\033f", next_word, NULL);
    add_keynode("M-b", "\033b", prev_word, NULL);
    add_keynode("M-bs", "\033\b", del_word_left, NULL);
    add_keynode("M-del", "\033\177", del_word_left, NULL);
    add_keynode("M-tab", "\033\t", complete_line, NULL);
}

#ifndef VT100
/*
 * extract the termcap 'cap' and put it in buf. Exit if critic != 0 and
 * 'cap' doesn't exist. Return 1 if 'cap' found, 0 if not.
 */
static int extract(cap, buf, critic)
char *cap, *buf;
int critic;
{
    char *p, b[80];
    
    p = b; *b = 0;
    if(!tgetstr(cap, &p)) {
	if(critic) {
	    fprintf(stderr,
		    "Your terminal %s is not powerful enough, missing '%s'.\n",
		    getenv("TERM"), cap);
	    exit(1);
	}
	*p = '\0';
	return 0;
    }
    /*
     * Remove the padding information. We assume that no terminals need
     * padding nowadays. At least it makes things much easier.
     */
    strcpy(buf, b + strspn(b, "0123456789*"));
    *p = '\0';
    return 1;
}
#endif /* VT100 */

/*
 * position the cursor using absolute coordinates
 * note: does not flush the output buffer
 */
void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf("\33[%d;%dH", col + 1, line + 1);
#else
  printf("%s", tgoto(curgoto, col, line));
#endif
}

/*
 * optimized cursor movement (no output flush)
 * from (fromline, fromcol) to (toline, tocol)
 */
void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[LINELEN];
    int cost = 0, i, dist;
    /*
     * First, move vertically to the correct line, then horisontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    
    if(fromline > toline) {
	if(*upcur) {
	    for(i = 0; i < (fromline - toline); i++) {
		strcpy(buf + cost, upcur);
		cost += strlen(upcur);
	    }
	} else {
	    /* we can't move cursor locally up - give up optimization */
	    gotoxy(tocol, toline);
	    return;
	}
    } else {
	cost = toline - fromline;
	for(i = 0; i < cost; i++) buf[i] = '\n';
	if(cost > 0) fromcol = 0; /* LF is mapped to CRLF on output */
    }
    if(fromcol > tocol) {
	dist = fromcol - tocol;
	if(strlen(leftcur) * dist <= strlen(begoln) + tocol) {
	    for(i = 0; i < dist; i++) {
		strcpy(buf + cost, leftcur);
		cost += strlen(leftcur);
	    }
	} else {
	    strcpy(buf + cost, begoln);
	    cost += strlen(begoln);
	    fromcol = 0;
	}
    }
    if(fromcol < tocol) {
	if(*edattrbeg) {
	    /* hiliting in effect: easy way out, just use gotoxy */
	    gotoxy(tocol, toline);
	    return;
	}
	if(toline == line0 && fromcol < col0) {
	    /* assume that if fromcol < col0, then tocol >= col0 */
	    strcpy(buf + cost, promptstr + fromcol);
	    cost += col0 - fromcol;
	    fromcol = col0;
	}
	strncpy(buf + cost, edbuf + (toline - line0) * (cols - 1) - col0
		+ fromcol, tocol - fromcol);
	cost += tocol - fromcol;
    }
    if(cost < gotocost) {
	buf[cost] = '\0';
	printf("%s", buf);
    } else
	gotoxy(tocol, toline);
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clearinpline()
{
    if(edlen) {
	optmove(CURLINE(pos), CURCOL(pos), line0, col0);
	printf("%s", clreoln);
	if(line0 < lines - 1) {
	    printf("\n%s", clreoscr);
	    optmove(line0 + 1, 0, line0, col0);
	}
	fflush(stdout);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    int i, oldline0;
    if(edlen) {
	oldline0 = line0;
	if(edlen < cols - col0 - 1) {
	    printf("%s%s", edattrbeg, edbuf);
	} else {
	    printf("%.*s", cols - col0 - 1, edbuf);
	    for(i = cols - col0 - 1; i <= edlen; i += cols - 1)
		printf("\n%.*s", cols - 1, edbuf + i);
	}
	line0 = lines - (edlen + col0) / (cols - 1) - 1;
	if(oldline0 < line0) line0 = oldline0;
	optmove(CURLINE(edlen), CURCOL(edlen), CURLINE(pos), CURCOL(pos));
    }
    fflush(stdout);
}

/*
 * delete n characters at current position (the position is unchanged)
 */
void deletechars(n)
int n;
{
  int i, j, cl, cc, endline;
  
  for(i = 0; i < n; i++) strdel(edbuf + pos);
  edlen -= n;
  
  /*
   * Write out the necessary changes. This code is a hell lot more
   * complicated than it need to be, and I am still not sure it works well
   * in all cases.
   */
  cc = CURCOL(pos); cl = CURLINE(pos);
  i = pos; endline = CURLINE(edlen);
  for(;;) 
   {
     if(*delchar) 
      {
	/* terminal can delete character, use it */
	j = (cols - 1 - cc < n) ? cols - 1 - cc : n;
	if(cl == endline && cc == CURCOL(edlen)) 
	 {
	   printf("%s", clreoln);
	 }
	else
	 {
	   while(j--) printf("%s", delchar);
	 }
	if(cl < endline) 
	 {
	   optmove(cl, cc, cl, cols - 1 - n);
	   i = (cl - line0) * (cols - 1) - col0 + cols - 1 - n;
	   printf("%.*s", n, edbuf + i);
	 }
	else
	   break;
      }
     else 
      {
	/* no delete character, alas */
	if(cl < endline) 
	 {
	   printf("%.*s", cols - 1 - cc, edbuf + i);
	   cc = cols - 1;
	 }
	else 
	 {
	   printf("%s%s", edbuf + i, clreoln);
	   cc = CURCOL(edlen);
	   break;
	 }
      }
     cl++; cc = 0;
     i = (cl - line0) * (cols - 1) - col0;
     putchar('\n');
   }
  optmove(cl, cc, CURLINE(pos), CURCOL(pos));
  fflush(stdout);
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insertchars(str, n)
char *str; int n;
{
  int i, j, k, cl, cc, endline;
  for(i = 0; i < n; i++) strins(edbuf + pos + i, str[i]);
  edlen += n;
  
  cc = CURCOL(pos); cl = CURLINE(pos); endline = CURLINE(edlen);
  i = pos;
  for(;;) 
   {
     if(*inschar) 
      {
	/*
	 * terminal can insert char, use it.
	 * The 3 is just ad-hockery: it is hard to calculate when inschar
	 * costs more than simply outputting the rest of the line, but
	 * it's not that important really.
	 */
	if(cl == endline && edlen - i < 3 * n) 
	 {
	   printf("%s", edbuf + i);
	   cc = CURCOL(edlen);
	 }
	else 
	 {
	   k = (cols - 1 - cc < n) ? cols - 1 - cc : n;
	   for(j = 0; j < k; j++) printf("%s", inschar);
	   printf("%.*s", k, edbuf + i);
	   cc += k;
	 }
	if(cl < endline) 
	 {
	   /*
	    * we inserted things into last column that isn't used, so
	    * we'd better delete that one to make everything look pretty.
	    */
	   optmove(cl, cc, cl, cols - 1);
	   printf("%s", clreoln);
	 }
	else
	   break;
      }
     else 
      {
	/* no inschar, print the whole line again */
	if(cl < endline) 
	 {
	   printf("%.*s", cols - 1 - cc, edbuf + i);
	   cc = cols - 1;
	 }
	else 
	 {
	   printf("%s", edbuf + i);
	   cc = CURCOL(edlen);
	   break;
	 }
      }
     cl++; cc = 0;
     if(cl == lines) 
      {
	cl--;
	line0--;
	endline--;
      }
     i = (cols - 1) * (cl - line0) - col0;
     putchar('\n');
   }
  pos += n;
  optmove(cl, cc, CURLINE(pos), CURCOL(pos));
  fflush(stdout);
}

/*
 * window size change signal handler:
 * we assume that the bottom line is always the bottom line
 */
void winchsig()
{
    struct winsize wsiz;
  
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &wsiz) == 0 && wsiz.ws_row && wsiz.ws_col) {
	lines = wsiz.ws_row;
	cols = wsiz.ws_col;
	send_term_size();
    }
    /* reinstall signal handler for system V unices */
    signal(SIGWINCH, winchsig);
}

